<?php


class Parcelforce_Calculator extends WPRuby_Abstract_Calculator
{


	public function __construct($settings)
	{
		$this->settings = $settings;
	}


	public function calculate($cart): array
	{

		$packages = $this->get_package_details_by_boxpacker($cart);
		$this->debug('Plugin Settings', $this->settings);
		$this->debug('Packing Details', $packages);
		if (!is_array($packages)) {
			return [];
		}

		return $this->get_price_for_packages($packages, $cart);
	}


	private function get_price_for_packages($packages, $cart)
	{

		$parcels = [];
		foreach ($packages as $package) {
			$parcels[] = [
				"Weight" => $package['weight'],
				"Length" => $package['length'],
				"Width" =>  $package['width'],
				"Height" => $package['height'],
			];
		}

		$countries = include ROYALMAILPRO_DIR . 'shared/data/countries.php';
		$countryName = WC()->countries->get_countries()[$cart['destination']['country']];
		$normalizedCountryName = $this->normalizeCountryName($countryName);
		if (!isset($countries[$normalizedCountryName])) {
			return [];
		}

		$body = [
			'CollectionCountry' => 219, // UK
			'CollectionPostcode' => '',
			'CollectionTown' => '',
			'DeliveryCountry' => intval($countries[$normalizedCountryName]),
			'DeliveryPostcode' => $cart['destination']['postcode'],
			'DeliveryTown' => $cart['destination']['city'],
			'Parcels' => $parcels,
			"QuoteType" => "",
			"MultiDeliveryDestinations" => false,
			"IsBusinessRecipient" => false,
			"QuoteModified" => false,
			"QuoteRequestType" => "Default",
			"ConfigurationProfile" => "PARCELFORCE.UK.LIVE"

		];

		$args = [
			'headers' => ['Content-Type' => 'application/json'],
			'body' => json_encode($body),
			'method' => 'POST',
			'timeout' => 10
		];

		$this->debug('Request', $body);

		$response = wp_remote_post('https://send.parcelforce.com/quotes/api/results/all', $args);

		if ($response instanceof WP_Error) {
			$this->debug('Response Error', $response->get_error_messages());
		}

		$responseBody = json_decode(wp_remote_retrieve_body($response), true);


		if (!( isset($responseBody['Quotes']) && is_array($responseBody['Quotes']) )) {
			return [];
		}

		$domestic_options = (!is_array($this->settings['domestic_delivery_options'])) ? [$this->settings['domestic_delivery_options'] ] : $this->settings['domestic_delivery_options'];

		$domestic_options = array_map(function ($service) {
			return sprintf('%s-%s-rec', $service, $this->settings['dispatch_type']);
		}, $domestic_options);

		$intl_options = (!is_array($this->settings['intl_delivery_options'])) ? [$this->settings['intl_delivery_options'] ] : $this->settings['intl_delivery_options'];

		$intl_options = array_map(function ($service) {
			return sprintf('%s-%s', $service, $this->settings['dispatch_type']);
		}, $intl_options);

		$enabledServices = array_merge($domestic_options, $intl_options);


		$totalCover = 0;
		if ($this->settings['add_cover'] === 'yes') {
			$totalCover = $this->getTotalCoverFee($packages, $cart['destination']['country'] === 'GB');
			$this->debug('Total Cover', $totalCover);
		}

		$rates = [];
		foreach ($responseBody['Quotes'] as $quote) {
			$serviceSlug = $quote['Service']['ServiceSlug'];

			if(in_array($serviceSlug, $enabledServices)) {
				$rate = [
					'id'    => $serviceSlug,
					'label' => sprintf('%s: %s', $this->settings['title'], $quote['Service']['HtmlName']),
					'cost'  => $quote['Total'] + $totalCover,
				];
				$rates[] = $rate;
				$this->debug('Rate', $rate);
			}

		}

		return $rates;

	}

	private function getTotalCoverFee($packages, $domestic = true)
	{
		if (!is_array($packages)) {
			return 0;
		}
		$total = 0;
		foreach ($packages as $package) {
			if ($domestic) {
				$total += $this->calculateDomesticCoverFees($package['value']);
			} else {
				$total += $this->calculateIntlCoverFees($package['value']);
			}
		}

		return $total;
	}

	/**
	 * @param float $value
	 * @return float
	 */
	private function calculateIntlCoverFees($value)
	{
		if ($value <= 100) {
			return 0;
		}

		if ($value <= 200 ) {
			return 1.5;
		}

		if ($value <= 300 ) {
			return 5.25;
		}

		if ($value <= 400 ) {
			return 9;
		}

		return 12.75;
	}


	/**
	 * @param float $value
	 * @return float
	 */
	private function calculateDomesticCoverFees($value)
	{
		if ($value <= 200) {
			return 0;
		}

		if ($value <= 300) {
			return 1.5;
		}

		// first 200 is free, the third hundred has value of 1.5.
		return ((($value / 100) - 3) * 3.75) + 1.5;
	}

}
